body {
    background-color: #eeeeee;
    color: black;
    text-align: left;
    font-family: "Myriad Set Pro", "Helvetica", sans-serif;
    font-size: 12px;
    margin-left: 10px
}


.headline {
    display: inline;
    font-size: 20px;
    margin-bottom: 20px;
}

.rect_icon {
    width: 20px;
    height: 20px;
    vertical-align: middle;
}

.headline_atom {
    position: absolute;
    top: 10px;
    left: 40px;
}

.table_margin_upper_table {
    margin-top: 20px;
    margin-left: -3px;
}

.p_margins {
    margin-top: 20px;
    margin-bottom: 20px;

}

.alertbox {
    border: 1px solid black;
    background-color: #ffeeee;
    padding: 10px;
    margin: 20px;
}

.alert {
    color: red;
    font-weight: bold;
}

.infobox {
    /* border: 1px solid black; */
    /* background-color: #ddeeff; */
    padding: 10px 0px;
    margin: 20px 0px;
}

.info {
    /* color: blue; */
    font-weight: bold;
}
