<!doctype html>
<html lang="en">
<head>
    <meta charset="UTF-8"/>

    <title>{{ title }}</title>
    <meta name="description" content="{{ description }}"/>
</head>

<body>

<div id="content">
    <div class="headline">Summary of Domain Surface-based Cavities</div>
    <table class="table_margin_upper_table">
        <tr>
            <td><strong>Domain Surface-based Cavities have not yet been calculated.</strong></td>
        </tr>
    </table>
</div>

</body>
</html>
